/*****************************************************************************
	MMPTP.C - Linux USB Kernel Module for PTP digital cameras

	(C) 2002 by Michael Minn - michael@michaelminn.com
	(C) 2002 Jos�Luis S�chez, wizord@mi.madritel.es

	printk messages are written to /var/log/messages

	See /usr/src/linux/Documentation/usb for more information
	about USB programming

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

*******************************************************************************/

#include <linux/modversions.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/usb.h>
#include <linux/init.h>
#include "mmptp.h"


/************************* Version Info *****************************/

#define DRIVER_VERSION "v2004.01.04"
#define DRIVER_AUTHOR "Michael Minn"
#define DRIVER_DESC "Linux USB Module for PTP digital cameras"


/************************* Globals *****************************/

static struct urb *mmptp_urb = NULL;
static int mmptp_in_endpoint = 0;
static int mmptp_out_endpoint = 0;
static int mmptp_interrupt_endpoint = 0;
static wait_queue_head_t mmptp_wait_queue;
static struct usb_device *mmptp_device = NULL;

static void mmptp_wait_callback(struct urb *urb)
{
	if (waitqueue_active(&mmptp_wait_queue))
		wake_up_interruptible(&mmptp_wait_queue);
}

/************************* Linux Device File Operations *****************************/

static int mmptp_file_open (struct inode *inode, struct file *file)
{
	/* int status = 0; */

	if (!mmptp_device)
		return -EHOSTDOWN;

	info("file opened");

	/*
	if ((status = usb_clear_halt(mmptp_device, usb_sndbulkpipe(mmptp_device, mmptp_out_endpoint))) != 0)
	{
		err("failure %d clearing stalled send pipe", status);
		return status;
	}

	if ((status = usb_clear_halt(mmptp_device, usb_rcvbulkpipe(mmptp_device, mmptp_in_endpoint))) != 0)
	{
		err("failure %d clearing stalled receive pipe", status);
		return status;
	}
	*/

	// MOD_INC_USE_COUNT;

	return 0;
}

static int mmptp_file_release (struct inode *inode, struct file *file)
{
	if (!mmptp_device)
		return -EHOSTDOWN;

	info("file closed");

	// MOD_DEC_USE_COUNT;

	return 0;
}

static ssize_t mmptp_file_write (struct file *file, const char *user_data, size_t length, loff_t *offset)
{
	int status = 0;

	/* info("write %d bytes", length); */

	if ((!mmptp_device) || (!mmptp_urb))
		return -EHOSTDOWN;

	mmptp_urb->status = 0;
	mmptp_urb->dev = mmptp_device;
	mmptp_urb->transfer_flags = 0;
	mmptp_urb->transfer_buffer_length = length;
	mmptp_urb->complete = mmptp_wait_callback;
	mmptp_urb->pipe = usb_sndbulkpipe(mmptp_device, mmptp_out_endpoint);
	copy_from_user(mmptp_urb->transfer_buffer, user_data, length);

	/*
	if ((status = usb_clear_halt(mmptp_device, usb_sndbulkpipe(mmptp_device, mmptp_out_endpoint))) != 0)
	{
		err("failure %d clearing stalled send pipe", status);
		return status;
	}
	*/

	if ((status = usb_submit_urb(mmptp_urb)) != 0)
	{
		err("failure %d submitting write urb", status);
		return status;
	}

	interruptible_sleep_on_timeout(&mmptp_wait_queue, HZ * 5);

	if (mmptp_urb->status == USB_ST_URB_PENDING)
	{
		err("Timed out waiting to send command %x", (int)((short*)mmptp_urb->transfer_buffer)[3]);
		usb_unlink_urb(mmptp_urb);
		return -ETIMEDOUT;
	}

	if (mmptp_urb->status == USB_ST_STALL)
	{
		if ((status = usb_clear_halt(mmptp_device, usb_rcvbulkpipe(mmptp_device, mmptp_in_endpoint))) != 0)
			err("failure %d clearing stalled receive pipe", status);

		if ((status = usb_clear_halt(mmptp_device, usb_sndbulkpipe(mmptp_device, mmptp_out_endpoint))) != 0)
			err("failure %d clearing stalled send pipe", status);

		info("Device indicated error by stalling endpoint while writing");
		return -EPIPE;
	}

	info("Write URB status %d", mmptp_urb->status);

	return length;
}

static ssize_t mmptp_file_read (struct file *file, char *data, size_t length, loff_t *offset)
{
	int status = 0;
	int current_length = 0;

	/* info("read %d bytes", length); */

	if ((!mmptp_device) || (!mmptp_urb))
		return -EHOSTDOWN;

	do
	{
		mmptp_urb->status = 0;
		mmptp_urb->dev = mmptp_device;
		mmptp_urb->transfer_flags = 0;
		mmptp_urb->transfer_buffer_length = 0x8000;
		/* mmptp_urb->transfer_buffer_length = 0x10000; */
		mmptp_urb->complete = mmptp_wait_callback;
		mmptp_urb->pipe = usb_rcvbulkpipe(mmptp_device, mmptp_in_endpoint);

		if ((status = usb_submit_urb(mmptp_urb)) != 0)
		{
			err("failure %d submitting read urb", status);
			return status;
		}

		interruptible_sleep_on_timeout(&mmptp_wait_queue, 5 * HZ);

		if (mmptp_urb->status == USB_ST_URB_PENDING)
		{
			err("Timed out waiting for response");
			usb_unlink_urb(mmptp_urb);
			return -ETIMEDOUT;
		}

		if (mmptp_urb->status == USB_ST_STALL)
		{
			unsigned short in_error = 0;
			unsigned short out_error = 0;

			/* This code does not work - when invalid object or storage is accessed */
			/* device stalls the endpoint.  I can't figure out how to clear stall errors */

			info("Device indicated error by stalling endpoint while reading");

			if ((status = usb_control_msg(mmptp_device, usb_rcvctrlpipe(mmptp_device, 0),
					USB_REQ_GET_STATUS, USB_DIR_IN | USB_RECIP_ENDPOINT, 0,
					mmptp_in_endpoint, &in_error, 2, HZ * 3)) != 2)
				err("Failure %d getting in endpoint status", status);

			if ((status = usb_control_msg(mmptp_device, usb_rcvctrlpipe(mmptp_device, 0),
					USB_REQ_GET_STATUS, USB_DIR_IN | USB_RECIP_ENDPOINT, 0,
					mmptp_out_endpoint, &out_error, 2, HZ * 3)) != 2)
				err("Failure %d getting out endpoint status", status);

			err("Endpoint error status: in %d, out %d", in_error, out_error);

			usb_unlink_urb(mmptp_urb);

			if ((status = usb_clear_halt(mmptp_device, usb_rcvbulkpipe(mmptp_device, mmptp_in_endpoint))) != 0)
				err("failure %d clearing in endpoint", status);
			if ((status = usb_clear_halt(mmptp_device, usb_sndbulkpipe(mmptp_device, mmptp_out_endpoint))) != 0)
				err("failure %d clearing out endpoint", status);
/*
			if ((status = usb_control_msg(mmptp_device, usb_sndctrlpipe(mmptp_device, 0),
					USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT, 0, mmptp_in_endpoint, NULL, 0, HZ * 3)) != 0)
				err("failure %d clearing in endpoint", status);

			if ((status = usb_control_msg(mmptp_device, usb_sndctrlpipe(mmptp_device, 0),
					USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT, 0, mmptp_out_endpoint, NULL, 0, HZ * 3)) != 0)
				err("failure %d clearing out endpoint", status);
*/

			return -EPIPE;
		}

		if ((current_length + mmptp_urb->actual_length) > length)
		{
			err("%d byte read() buffer too small for %d byte message", length, mmptp_urb->transfer_buffer_length);
			mmptp_urb->transfer_buffer_length = length;
		}

	/* info("%d bytes received", mmptp_urb->transfer_buffer_length); */

		copy_to_user(&data[current_length], mmptp_urb->transfer_buffer, mmptp_urb->actual_length);
		current_length += mmptp_urb->actual_length;
	}
	while (mmptp_urb->actual_length >= mmptp_urb->transfer_buffer_length);

	return current_length;
}

static int mmptp_file_ioctl (struct inode *inode, struct file *file, unsigned int command, unsigned long data)
{
	info("ioctl %x", command);

	return -EINVAL;
}

struct file_operations mmptp_file_operations =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* struct module *owner; */
#endif
	NULL, /* loff_t (*llseek) (struct file *, loff_t, int); */
	mmptp_file_read, /* ssize_t (*read) (struct file *, char *, size_t, loff_t *); */
	mmptp_file_write, /* ssize_t (*write) (struct file *, const char *, size_t, loff_t *); */
	NULL, /* int (*readdir) (struct file *, void *, filldir_t); */
	NULL, /* unsigned int (*poll) (struct file *, struct poll_table_struct *); */
	mmptp_file_ioctl, /* int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); */
	NULL, /* int (*mmap) (struct file *, struct vm_area_struct *); */
	mmptp_file_open, /* int (*open) (struct inode *, struct file *); */
	NULL, /* int (*flush) (struct file *); */
	mmptp_file_release, /* int (*release) (struct inode *, struct file *); */
	NULL, /* int (*fsync) (struct file *, struct dentry *); */
	NULL, /* int (*fasync) (int, struct file *, int); */
	NULL, /* int (*lock) (struct file *, int, struct file_lock *); */
	NULL, /* ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *); */
	NULL, /* ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *); */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* ssize_t (*writepage) (struct file *, struct page *, int, size_t, loff_t *, int); */
#endif
};


/************************* Debugging *****************************/

static void mmptp_list_descriptors(struct usb_device *device)
{
	int configuration;
	for (configuration = 0; configuration < device->descriptor.bNumConfigurations; ++configuration)
	{
		int interface;
		info("Configuration %d, %d interfaces", configuration, device->config[configuration].bNumInterfaces);

		for (interface = 0; interface < device->config[configuration].bNumInterfaces; ++interface)
		{
			int alternate;
			info("Interface %d, %d alternates", interface, device->config[configuration].interface[interface].num_altsetting);
			for (alternate = 0; alternate < device->config[configuration].interface[interface].num_altsetting; ++alternate)
			{
				int endpoint;
				info("Alternate %d, %d endpoints", interface, device->config[configuration].interface[interface].altsetting[alternate].bNumEndpoints);
				for (endpoint = 0; endpoint < device->config[configuration].interface[interface].altsetting[alternate].bNumEndpoints; ++endpoint)
				{
					info("Endpoint %d: address %x, max packet %d, ",
						endpoint, device->config[configuration].interface[interface].altsetting[alternate].endpoint[endpoint].bEndpointAddress,
						device->config[configuration].interface[interface].altsetting[alternate].endpoint[endpoint].wMaxPacketSize);

					switch (device->config[configuration].interface[interface].altsetting[alternate].endpoint[endpoint].bmAttributes)
					{
						case USB_ENDPOINT_XFER_CONTROL: info("Control"); break;
						case USB_ENDPOINT_XFER_ISOC: info("Isosynchronous"); break;
						case USB_ENDPOINT_XFER_BULK: info("Bulk"); break;
						case USB_ENDPOINT_XFER_INT: info("Interrupt"); break;
					}
				}
			}
		}
	}
}

/************************* USB Module Registration, Entry, Exit *****************************/


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
static void *mmptp_probe(struct usb_device *device, unsigned interface, const struct usb_device_id *id)
#else
static void *mmptp_probe(struct usb_device *device, unsigned interface)
#endif
{
	if (device &&
		 device->config &&
		 device->config->interface &&
		 device->config->interface->altsetting &&
		 (device->config->interface->altsetting->bInterfaceSubClass == 0x01) &&
		 (device->config->interface->altsetting->bInterfaceProtocol == 0x01))
	{
		info("PTP Class Imaging Device Connected - Support not guaranteed");
		mmptp_list_descriptors(device);

		mmptp_device = device;
		mmptp_urb = usb_alloc_urb(0);
		mmptp_urb->transfer_buffer = (unsigned char *)kmalloc(0x8000, GFP_KERNEL);

		mmptp_in_endpoint = 0x81;
		mmptp_out_endpoint = 0x01;
		mmptp_interrupt_endpoint = 0x82;

		return device;
	}

	else if ((device->descriptor.idVendor == 0x40a) && (device->descriptor.idProduct == 0x500) && (interface == 0))
	{
		info("Kodak DX-3500 Connected");
		mmptp_list_descriptors(device);

		mmptp_device = device;
		mmptp_urb = usb_alloc_urb(0);
		mmptp_urb->transfer_buffer = (unsigned char *)kmalloc(0x8000, GFP_KERNEL);

		mmptp_in_endpoint = 0x81;
		mmptp_out_endpoint = 0x01;
		mmptp_interrupt_endpoint = 0x82;

		return device;
	}

	else if ((device->descriptor.idVendor == 0x40a) && (device->descriptor.idProduct == 0x160) && (interface == 0))
	{
		info("Kodak DC-4800 Connected - Support not guaranteed");
		mmptp_list_descriptors(device);

		mmptp_device = device;
		mmptp_urb = usb_alloc_urb(0);
		mmptp_urb->transfer_buffer = (unsigned char *)kmalloc(0x8000, GFP_KERNEL);

		mmptp_in_endpoint = 0x81;
		mmptp_out_endpoint = 0x01;
		mmptp_interrupt_endpoint = 0x82;

		return device;
	}

	else if ((device->descriptor.idVendor == 0x40a) && (device->descriptor.idProduct == 0x121) && (interface == 0))
	{
		info("Kodak DC-240 Connected - Support not guaranteed");
		mmptp_list_descriptors(device);

		mmptp_device = device;
		mmptp_urb = usb_alloc_urb(0);
		mmptp_urb->transfer_buffer = (unsigned char *)kmalloc(0x8000, GFP_KERNEL);

		mmptp_in_endpoint = 0x81;
		mmptp_out_endpoint = 0x01;
		mmptp_interrupt_endpoint = 0x82;

		return device;
	}

	return NULL;
}

static void mmptp_disconnect(struct usb_device *device, void *driver_context)
{
	if (mmptp_urb)
	{
		usb_unlink_urb(mmptp_urb);
		kfree(mmptp_urb->transfer_buffer);
		usb_free_urb(mmptp_urb);
		mmptp_urb = NULL;
	}

	mmptp_device = NULL;

	info("Kodak DX3500 Disconnected");
}

static struct usb_driver mmptp_driver_registration =
{
	name: "mmptp",
	probe: mmptp_probe,
	disconnect: mmptp_disconnect,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	id_table:	NULL,
	driver_list: {NULL, NULL},
	fops: NULL
#endif
};

// int init_module(void)
static int __init mmptp_init(void)
{
	int status = 0;

	info(DRIVER_DESC " " DRIVER_VERSION);
	info("Major Number %d", MMPTP_MAJOR_NUMBER);

	init_waitqueue_head(&mmptp_wait_queue);

	if ((status = register_chrdev(MMPTP_MAJOR_NUMBER, "mmptp", &mmptp_file_operations)) < 0)
	{
		err("register_chrdev() failed code %d", status);
		return -1;
	}

	if ((status = usb_register(&mmptp_driver_registration)) < 0)
		err("Driver failed usb_register with code %d", status);

	return 0;
}

// int cleanup_module(void)
static void __exit mmptp_cleanup(void)
{
	unregister_chrdev(MMPTP_MAJOR_NUMBER, "mmptp");

	usb_deregister(&mmptp_driver_registration);

	info("Module deregistered and unloaded");

	// return 0;
}

module_init(mmptp_init);
module_exit(mmptp_cleanup);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif



